"""A package for sleep stage classification using ECG data."""

from .config import get_config, set_config
from .feature_extraction import extract_features, preprocess_rri
from .heartbeats import compare_heartbeats, detect_heartbeats, rri_similarity
from .io import *  # noqa: F403

__version__ = '0.4.1'



# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""Functions for reading datasets containing ECG and beat annotations."""

from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, Optional, Union

import numpy as np
import requests
from tqdm import tqdm

from ..config import get_config
from .physionet import _list_physionet, download_physionet
from .utils import _download_file


@dataclass
class ECGRecord:
    """
    Dataclass to store a single ECG record.

    Attributes
    ----------
    ecg : np.ndarray
        The ECG signal.
    fs : float
        The sampling frequency.
    annotation : np.ndarray
        Indices of annotated heartbeats.
    lead : str, optional
        Which ECG lead the signal was recorded from, by default `None`.
    id : str, optional
        The record's ID, by default `None`.
    """

    ecg: np.ndarray
    fs: float
    annotation: np.ndarray
    lead: Optional[str] = None
    id: Optional[str] = None


def read_ltdb(
    records_pattern: str = '*',
    offline: bool = False,
    data_dir: Optional[Union[str, Path]] = None,
) -> Iterator[ECGRecord]:
    """
    Lazily read records from LTDB (https://physionet.org/content/ltdb/).

    Parameters
    ----------
    records_pattern : str, optional
        Glob-like pattern to select record IDs, by default `'*'`.
    offline : bool, optional
        If `True`, only local files will be used (i.e. no files will be
        downloaded), by default `False`.
    data_dir : str | pathlib.Path, optional
        Directory where all datasets are stored. If `None` (default), the
        value will be taken from the configuration.

    Yields
    ------
    ECGRecord
        Each element in the generator is of type `ECGRecord` and contains
        the ECG signal (`.ecg`), sampling frequency (`.fs`), annotated beat
        indices (`.annotations`), `.lead`, and `.id`.
    """
    if data_dir is None:
        data_dir = get_config('data_dir')
    yield from _read_mitbih('ltdb', records_pattern, offline, data_dir)


def read_mitdb(
    records_pattern: str = '*',
    offline: bool = False,
    data_dir: Optional[Union[str, Path]] = None,
) -> Iterator[ECGRecord]:
    """
    Lazily read records from MITDB (https://physionet.org/content/mitdb/).

    Parameters
    ----------
    records_pattern : str, optional
        Glob-like pattern to select record IDs, by default `'*'`.
    offline : bool, optional
        If `True`, only local files will be used (i.e. no files will be
        downloaded), by default `False`.
    data_dir : str | pathlib.Path, optional
        Directory where all datasets are stored. If `None` (default), the
        value will be taken from the configuration.

    Yields
    ------
    ECGRecord
        Each element in the generator is of type `ECGRecord` and contains
        the ECG signal (`.ecg`), sampling frequency (`.fs`), annotated beat
        indices (`.annotations`), `.lead`, and `.id`.
    """
    if data_dir is None:
        data_dir = get_config('data_dir')
    yield from _read_mitbih('mitdb', records_pattern, offline, data_dir)


def _read_mitbih(
    db_slug: str,
    records_pattern: str,
    offline: bool,
    data_dir: Union[str, Path],
) -> Iterator[ECGRecord]:
    """
    Lazily reads records from MIT-BIH datasets (e.g. MITDB, LTDB).

    Required files are downloaded if not present in `<data_dir>/<db_slug>`.

    Parameters
    ----------
    db_slug : str
        Short identifier of a database, e.g. `'mitdb'`.
    records_pattern : str
        Glob-like pattern to select record IDs.
    offline : bool
        If `True`, only local files will be used (i.e. no files will be
        downloaded).
    data_dir : str | pathlib.Path
        Directory where all datasets are stored.

    Yields
    ------
    ECGRecord
        Each element in the generator is of type `ECGRecord` and contains
        the ECG signal (`.ecg`), sampling frequency (`.fs`), annotated beat
        indices (`.annotations`), `.lead`, and `.id`.
    """
    import wfdb

    # https://archive.physionet.org/physiobank/database/html/mitdbdir/intro.htm#symbols
    BEAT_ANNOTATION_SYMBOLS = set('NLRAaJSFejE/fQ|')

    data_dir = Path(data_dir).expanduser()

    requested_records = _list_physionet(data_dir, db_slug, pattern=records_pattern)

    if not offline:
        download_physionet(
            data_dir,
            db_slug,
            requested_records,
            extensions=['.hea', '.dat', '.atr'],
        )

    for record_id in requested_records:
        record_file = str(data_dir / db_slug / record_id)

        annotations = wfdb.rdann(record_file, 'atr')
        beat_indices = []
        for sample, symbol in zip(annotations.sample, annotations.symbol):
            if symbol in BEAT_ANNOTATION_SYMBOLS:
                beat_indices.append(sample)

        record = wfdb.rdrecord(record_file)

        for signal_index, signal_name in enumerate(record.sig_name):
            yield ECGRecord(
                ecg=record.p_signal[:, signal_index].flatten(),
                fs=record.fs,
                annotation=np.array(beat_indices, dtype=int),
                lead=signal_name,
                id=record_id,
            )


def read_gudb(
    offline: bool = False,
    data_dir: Optional[Union[str, Path]] = None,
) -> Iterator[ECGRecord]:
    """
    Lazily reads records from GUDB (https://berndporr.github.io/ECG-GUDB/).

    Required files are downloaded if not present in `'<data_dir>/gudb'`.

    Parameters
    ----------
    offline : bool, optional
        If `True`, only local files will be used (i.e. no files will be
        downloaded), by default `False`.
    data_dir : str | pathlib.Path, optional
        Directory where all datasets are stored. If `None` (default), the
        value will be taken from the configuration.

    Yields
    ------
    ECGRecord
        Each element in the generator is of type `ECGRecord` and contains
        the ECG signal (`.ecg`), sampling frequency (`.fs`), annotated beat
        indices (`.annotations`), `.lead`, and `.id`.
    """
    import pandas as pd

    DB_URL = 'https://berndporr.github.io/ECG-GUDB/experiment_data'
    EXPERIMENTS = ['sitting', 'maths', 'walking', 'hand_bike', 'jogging']
    FS = 250

    if data_dir is None:
        data_dir = get_config('data_dir')

    db_dir = Path(data_dir).expanduser() / 'gudb'

    for subject_id in tqdm(list(range(25)), desc='Reading GUDB'):
        for experiment in EXPERIMENTS:
            experiment_subdir = f'subject_{subject_id:02}/{experiment}'
            if not offline:
                for tsv_filename in (
                    'ECG.tsv',
                    'annotation_cs.tsv',
                    'annotation_cables.tsv',
                ):
                    ecg_file_url = f'{DB_URL}/{experiment_subdir}/{tsv_filename}'
                    target_filepath = db_dir / experiment_subdir / tsv_filename
                    try:
                        _download_file(ecg_file_url, target_filepath)
                    except requests.exceptions.HTTPError as error:
                        print(error)
            ecg_data = pd.read_csv(
                db_dir / experiment_subdir / 'ECG.tsv',
                sep=' ',  # contrary to what .tsv suggests, the data is space-separated
                names=['chest', 'II', 'III', 'X', 'Y', 'Z'],
            )
            annotations_chest_file = db_dir / experiment_subdir / 'annotation_cs.tsv'
            if annotations_chest_file.is_file():
                yield ECGRecord(
                    ecg=ecg_data['chest'].to_numpy(),
                    fs=FS,
                    annotation=np.loadtxt(annotations_chest_file, dtype=np.int32),
                    lead='chest',
                    id=f'{subject_id:02}_{experiment}',
                )
            annotations_chest_file = db_dir / experiment_subdir / 'annotation_cables.tsv'
            if annotations_chest_file.is_file():
                annotations = np.loadtxt(annotations_chest_file, dtype=np.int32)
                for lead in ('II', 'III'):
                    yield ECGRecord(
                        ecg=ecg_data[lead].to_numpy(),
                        fs=FS,
                        annotation=annotations,
                        lead=lead,
                        id=f'{subject_id:02}_{experiment}',
                    )



# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""Functions and utilities related to feature extraction."""

import warnings
from typing import Iterable, List, Optional, Tuple, Union

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
from scipy.interpolate import interp1d
from scipy.signal import periodogram

from .io.sleep_readers import SleepRecord
from .utils import _parallel

_FEATURE_GROUPS = {
    'hrv-time': (
        'meanNN', 'maxNN', 'minNN', 'rangeNN', 'SDNN',
        'RMSSD', 'SDSD', 'NN50', 'NN20', 'pNN50', 'pNN20',
        'medianNN', 'madNN', 'iqrNN', 'cvNN', 'cvSD',
        'meanHR', 'maxHR', 'minHR', 'stdHR',
        'SD1', 'SD2', 'S', 'SD1_SD2_ratio', 'CSI', 'CVI',
    ),
    'hrv-frequency': (
        'total_power', 'VLF', 'LF', 'LF_norm', 'HF', 'HF_norm', 'LF_HF_ratio',
    ),
}
_FEATURE_ID_TO_GROUP = {id: group for group, ids in _FEATURE_GROUPS.items() for id in ids}


def _create_ragged_array(data: List[np.ndarray]) -> np.ndarray:
    """
    Convert a list of arrays with different lengths to a numpy array.

    Each element in `data` is a row in the resulting array. Rows shorter
    than the longest row will be padded with `np.nan`.

    Parameters
    ----------
    data : list[np.ndarray]
        A list of arrays which may have different lengths.

    Returns
    -------
    np.ndarray
        The padded rectangular array.
    """
    max_len = max(len(x) for x in data)
    ragged_array = np.full((len(data), max_len), fill_value=np.nan)
    for row_index, row in enumerate(data):
        ragged_array[row_index, :len(row)] = row
    return ragged_array


def _split_into_windows(
    data: np.ndarray,
    data_times: np.ndarray,
    window_times: np.ndarray,
    lookback: int,
    lookforward: int,
) -> List[np.ndarray]:
    """
    Split (irregularly sampled) data into windows of equal temporal length.

    Make sure `data_times`, `window_times`, `lookback` and `lookforward`
    use the same units.

    Parameters
    ----------
    data : np.ndarray
        The data to split.
    data_times : np.ndarray
        Sampling times of `data`.
    window_times : np.ndarray
        Times at which windows should be created.
    lookback : int
        Backward extension of the window (i.e. "to the left").
    lookforward : int
        Forward extension of the window (i.e. "to the left").

    Returns
    -------
    list[np.ndarray]
        A list containing each window as an array. Note that each window
        may contain a different number of elements in case the data is
        sampled irregularly.
    """
    window_start_times = window_times - lookback
    window_end_times = window_times + lookforward

    windows = []
    for start, end in zip(window_start_times, window_end_times):
        windows.append(data[(start <= data_times) & (data_times < end)])

    return windows


def _nanpsd(x: np.ndarray, fs: float, max_nans: float = 0) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute power spectral density (PSD) along axis 1, ignoring NaNs.

    For rows containing a fraction of NaNs higher than `max_nans`, the
    output array `Pxx` is filled with `np.nan`.

    Parameters
    ----------
    x : np.ndarray
        2d array where each row is treated as an individual signal.
    fs : float
        Sampling frequency in Hz.
    max_nans : float, optional
        Maximum fraction of NaNs in a signal (i.e. row of `x`), for which
        the PSD computation is attempted. Should be a value between `0.0`
        and `1.0`, by default `0`.

    Returns
    -------
    f : np.ndarray
        Array of sample frequencies.
    Pxx : np.ndarray
        One-sided power spectral density of x.
    """
    nfft = x.shape[1]
    Pxx = np.full((x.shape[0], nfft // 2 + 1), np.nan)

    nan_fraction = np.mean(np.isnan(x), axis=1)

    # rows without any NaNs
    full_rows_mask = nan_fraction == 0
    f, Pxx[full_rows_mask] = periodogram(x=x[full_rows_mask], fs=fs)

    # remaining rows with less than max_nans NaNs
    empty_rows_mask = nan_fraction == 1
    for i in np.where((nan_fraction <= max_nans) & ~(full_rows_mask | empty_rows_mask))[0]:
        semi_valid_window = x[i]
        valid_part = semi_valid_window[~np.isnan(semi_valid_window)]
        _, Pxx[i] = periodogram(valid_part, fs=fs, window='hann', nfft=nfft)

    return f, Pxx


def _hrv_timedomain_features(
    rri: np.ndarray,
    rri_times: np.ndarray,
    stage_times: np.ndarray,
    lookback: int,
    lookforward: int,
) -> np.ndarray:
    """
    Calculate 26 time domain heart rate variability (HRV) features.

    Features are implemented according to [1]_, [2]_ and [3]_.

    Parameters
    ----------
    rri : np.ndarray
        1d-array containing RR-intervals in seconds.
    rri_times : np.ndarray
        1d-array containing sample times of `rri` in seconds.
    stage_times : np.ndarray
        1d-array containing sleep stage onset times in seconds.
    lookback : int
        Backward extension of the analysis window from each sleep stage
        time.
    lookforward : int
        Forward extension of the analysis window from each sleep stage
        time.

    Returns
    -------
    np.ndarray
        Array of shape `(len(stage_times), 26)` containing the extracted
        time domain features.

    Notes
    -----
    .. [1] Task Force of the European Society of Cardiology. (1996). Heart
       rate variability: standards of measurement, physiological
       interpretation and clinical use. circulation, 93, 1043-1065.
       https://doi.org/10.1161/01.CIR.93.5.1043
    .. [2] Shaffer, F., & Ginsberg, J. P. (2017). An overview of heart rate
       variability metrics and norms. Frontiers in public health, 258.
       https://doi.org/10.3389/fpubh.2017.00258
    .. [3] Toichi, M., Sugiura, T., Murai, T., & Sengoku, A. (1997). A new
       method of assessing cardiac autonomic function and its comparison
       with spectral analysis and coefficient of variation of R–R interval.
       Journal of the autonomic nervous system, 62(1-2), 79-84.
       https://doi.org/10.1016/S0165-1838(96)00112-9
    """
    NN = _split_into_windows(
        rri,
        rri_times,
        stage_times,
        lookback,
        lookforward,
    )
    NN = _create_ragged_array(NN)

    meanNN = np.nanmean(NN, axis=1)
    maxNN = np.nanmax(NN, axis=1)
    minNN = np.nanmin(NN, axis=1)
    rangeNN = maxNN - minNN
    SDNN = np.nanstd(NN, axis=1, ddof=1)

    SD = np.diff(NN)
    RMSSD = np.sqrt(np.nanmean(SD**2, axis=1))
    SDSD = np.nanstd(SD, axis=1, ddof=1)
    NN50 = np.nansum(np.abs(SD) > 0.05, axis=1)
    NN20 = np.nansum(np.abs(SD) > 0.02, axis=1)
    pNN50 = np.nanmean(np.abs(SD) > 0.05, axis=1)
    pNN20 = np.nanmean(np.abs(SD) > 0.02, axis=1)

    medianNN = np.nanmedian(NN, axis=1)
    madNN = np.nanmedian(np.abs(NN - medianNN[:, np.newaxis]), axis=1)
    iqrNN = np.nanpercentile(NN, 75, axis=1) - np.nanpercentile(NN, 25, axis=1)

    cvNN = SDNN / meanNN
    cvSD = SDSD / np.nanmean(SD, axis=1)

    meanHR = 60 / meanNN
    maxHR = 60 / minNN
    minHR = 60 / maxNN
    stdHR = np.nanstd(60 / NN, axis=1, ddof=1)

    SD1 = (SDSD**2 * 0.5)**0.5
    SD2 = (2 * SDNN**2 - SD1**2)**0.5
    S = np.pi * SD1 * SD2
    SD1_SD2_ratio = SD1 / SD2

    CSI = SD2 / SD1
    CVI = np.log10(SD1 * SD2 * 16)

    return np.vstack((
        meanNN, maxNN, minNN, rangeNN, SDNN,
        RMSSD, SDSD, NN50, NN20, pNN50, pNN20,
        medianNN, madNN, iqrNN, cvNN, cvSD,
        meanHR, maxHR, minHR, stdHR,
        SD1, SD2, S, SD1_SD2_ratio, CSI, CVI,
    )).T


def _hrv_frequencydomain_features(
    rri: np.ndarray,
    rri_times: np.ndarray,
    stage_times: np.ndarray,
    lookback: int,
    lookforward: int,
    fs_rri_resample: float,
    max_nans: float,
    feature_ids: List[str],
) -> np.ndarray:
    """
    Calculate seven frequency domain heart rate variability (HRV) features.

    Features are implemented according to [1]_.

    Parameters
    ----------
    rri : np.ndarray
        1d-array containing RR-intervals in seconds.
    rri_times : np.ndarray
        1d-array containing sample times of `rri` in seconds.
    stage_times : np.ndarray
        1d-array containing sleep stage onset times in seconds. Distances
        between onsets must be regular.
    lookback : int
        Backward extension of the analysis window from each sleep stage
        time.
    lookforward : int
        Forward extension of the analysis window from each sleep stage
        time.
    fs_rri_resample : float
        Frequency in Hz at which the RRI time series should be resampled
        before spectral analysis.
    max_nans : float
        Maximum fraction of NaNs in an analysis window, for which frequency
        features are computed. Should be a value between `0.0` and `1.0`.
    feature_ids : list[str]
        A list containing the identifiers of all features to be extracted.
        This does not change the returned array. It is only used to avoid
        issuing a warning about the analysis window being too short for
        some frequency range which isn't requested.

    Returns
    -------
    np.ndarray
        Array of shape `(len(stage_times), 7)` containing the extracted
        frequency domain features.

    Notes
    -----
    .. [1] Task Force of the European Society of Cardiology. (1996). Heart
       rate variability: standards of measurement, physiological
       interpretation and clinical use. circulation, 93, 1043-1065.
       https://doi.org/10.1161/01.CIR.93.5.1043
    """
    # The recording should last for at least 10 times the wavelength of the
    # lower frequency bound of the investigated component.
    window_time = lookback + lookforward
    min_frequencies = {
        'VLF': 0.0033,
        'LF': 0.04,
        'LF_norm': 0.04,
        'HF': 0.15,
        'HF_norm': 0.04,
        'LF_HF_ratio': 0.04,
    }

    for name, min_frequency in min_frequencies.items():
        min_window_time = 10 / min_frequency
        if name not in feature_ids:
            continue
        if window_time < min_window_time:
            msg = (
                f'HR analysis window too short for estimating PSD for feature {name}. '
                f'{min_window_time:.1f}s required, got {window_time}s'
            )
            warnings.warn(msg, category=RuntimeWarning)

    rri_interp_times = np.arange(
        start=stage_times[0] - lookback,
        stop=stage_times[-1] + lookforward,
        step=1 / fs_rri_resample,
    )
    rri_interp = interp1d(rri_times, rri, bounds_error=False)(rri_interp_times)

    # create (overlapping) windows, 1 per sleep stage
    sleep_stage_durations = np.diff(stage_times)
    if np.any(sleep_stage_durations != sleep_stage_durations[0]):
        raise ValueError('Sleep stages must be sampled regularly!')

    window_size = (lookback + lookforward) * fs_rri_resample
    window_step = int(fs_rri_resample * sleep_stage_durations[0])
    rri_windows = sliding_window_view(rri_interp, window_size)[::window_step]

    freq, psd = _nanpsd(rri_windows, fs_rri_resample, max_nans)

    total_power_mask = freq <= 0.4
    vlf_mask = (0.0033 < freq) & (freq <= 0.04)
    lf_mask = (0.04 < freq) & (freq <= 0.15)
    hf_mask = (0.15 < freq) & (freq <= 0.4)

    total_power = np.trapz(psd[:, total_power_mask], freq[total_power_mask])
    vlf = np.trapz(psd[:, vlf_mask], freq[vlf_mask])
    lf = np.trapz(psd[:, lf_mask], freq[lf_mask])
    hf = np.trapz(psd[:, hf_mask], freq[hf_mask])

    lf_norm = lf / (lf + hf) * 100
    hf_norm = hf / (lf + hf) * 100
    lf_hf_ratio = lf / hf

    return np.vstack((total_power, vlf, lf, lf_norm, hf, hf_norm, lf_hf_ratio)).T


def _parse_feature_selection(
    requested_ids: List[str],
) -> Tuple[List[str], List[str], List[int]]:
    """
    Parse a list containing feature group names and single feature IDs.

    Each feature group is expanded to all its feature identifiers as listed
    in `feature_extraction._FEATURE_GROUPS`, preserving input order. If an
    invalid (group) ID is found, a `ValueError` is raised.

    Parameters
    ----------
    requested_ids : list[str]
        A list which can contain both feature group names and single
        feature IDs in arbitrary order.

    Returns
    -------
    required_groups : list[str]
        The feature groups which have to be calculated to cover all
        requested features.
    feature_ids : list[str]
        The expanded list containing only single feature IDs.
    selected_cols : list[int]
        The column indices of `feature_ids` in a list of all feature IDs in
        all `required_groups`.
    """
    required_groups = set()
    feature_ids = []

    for id_ in requested_ids:
        if id_ in _FEATURE_GROUPS:
            required_groups.add(id_)
            feature_ids.extend(_FEATURE_GROUPS[id_])
        elif id_ in _FEATURE_ID_TO_GROUP:
            required_groups.add(_FEATURE_ID_TO_GROUP[id_])
            feature_ids.append(id_)
        else:
            raise ValueError(f'Invalid feature or group ID: {id_}')

    all_cols = [id for group in required_groups for id in _FEATURE_GROUPS[group]]
    selected_cols = [all_cols.index(id_) for id_ in feature_ids]

    duplicate_ids = {x for x in feature_ids if feature_ids.count(x) > 1}
    if duplicate_ids:
        warnings.warn(f'Duplicates in feature selection: {duplicate_ids}', RuntimeWarning)

    return list(required_groups), feature_ids, selected_cols


def preprocess_rri(
    rri: np.ndarray,
    min_rri: Optional[float] = None,
    max_rri: Optional[float] = None,
):
    """
    Replace invalid RRI samples with `np.nan`.

    Parameters
    ----------
    rri : np.ndarray
        An array containing consecutive RR interval lengths in seconds.
    min_rri : float, optional
        Minimum RRI in seconds to be considered valid. If `None` (default),
        no lower bounds check is performed.
    max_rri : float, optional
        Maximum RRI in seconds to be considered valid. If `None` (default),
        no upper bounds check is performed.

    Returns
    -------
    np.ndarray
        The cleaned RRI series.
    """
    invalid_rri = np.zeros_like(rri, dtype=bool)
    if min_rri is not None:
        invalid_rri |= rri < min_rri
    if max_rri is not None:
        invalid_rri |= rri > max_rri
    rri[invalid_rri] = np.nan
    return rri


def _extract_features_single(
    record: SleepRecord,
    sleep_stage_duration: int,
    min_rri: float,
    max_rri: float,
    required_groups: List[str],
    lookback: int,
    lookforward: int,
    fs_rri_resample: float,
    max_nans: float,
    feature_ids: List[str],
    col_indices: List[int],
) -> Tuple[np.ndarray, Union[np.ndarray, None]]:
    """
    Calculate features for a single record.

    This function is required to allow parallelizing feature extraction.

    Parameters
    ----------
    record : SleepRecord
        The record for which to calculate features.
    sleep_stage_duration : int
        Duration of a single sleep stage in the returned `stages` in
        seconds.
    min_rri: float, optional
        Minimum RRI value in seconds to be considered valid. Will be passed
        to :func:`preprocess_rri`.
    max_rri: float, optional
        Maximum RRI value in seconds to be considered valid. Will be passed
        to :func:`preprocess_rri`.
    required_groups : list[str]
        The feature groups which have to be calculated to cover all
        requested features.
    lookback : int, optional
        Backward extension of the analysis window from each sleep stage
        time in seconds.
    lookforward : int, optional
        Forward extension of the analysis window from each sleep stage
        time in seconds.
    fs_rri_resample : float
        Frequency in Hz at which the RRI time series should be resampled
        before spectral analysis. Only relevant for frequency domain
        features.
    max_nans : float
        Maximum fraction of NaNs in an analysis window for which frequency
        features are computed. Should be a value between `0.0` and `1.0`.
    feature_ids : list[str]
        A list containing the identifiers of all features to be extracted.
        This is only used to avoid issuing a warning about the analysis
        window being too short for some frequency range which isn't
        requested.
    col_indices : list[int]
        The column indices of `feature_ids` in a list of all feature IDs in
        all `required_groups`. Required to select the columns corresponding
        to the requested `feature_ids` from the calculated feature matrix.

    Returns
    -------
    features : np.ndarray
        The feature matrix of shape `(len(sleep_stages), <num_features>)`
        containing the extracted features.
    stages : np.ndarray | None
        The label vector, i.e. the annotated sleep stages. For a `record`
        without annotated stages, this will be `None`.
    """
    rri_required = 'hrv-time' in required_groups or 'hrv-frequency' in required_groups

    if record.sleep_stages is not None and record.sleep_stage_duration is not None:
        record_duration = len(record.sleep_stages) * record.sleep_stage_duration
    elif record.heartbeat_times is not None:
        record_duration = record.heartbeat_times[-1]
    else:
        raise ValueError(f'Record duration cannot be inferred for {record.id}.')
    num_stages = record_duration // sleep_stage_duration
    stage_times = np.arange(num_stages) * sleep_stage_duration

    if rri_required:
        if record.heartbeat_times is None:
            raise ValueError(f'Cannot extract HRV features for record {record.id} without heartbeat_times.')  # noqa: E501
        rri = preprocess_rri(
            np.diff(record.heartbeat_times),
            min_rri=min_rri,
            max_rri=max_rri,
        )
        rri_times = record.heartbeat_times[1:]

    X = []
    for feature_group in required_groups:
        if feature_group == 'hrv-time':
            X.append(
                _hrv_timedomain_features(
                    rri,
                    rri_times,
                    stage_times,
                    lookback,
                    lookforward,
                ),
            )
        elif feature_group == 'hrv-frequency':
            X.append(
                _hrv_frequencydomain_features(
                    rri,
                    rri_times,
                    stage_times,
                    lookback,
                    lookforward,
                    fs_rri_resample,
                    max_nans,
                    feature_ids,
                ),
            )
    features = np.hstack(X)[:, col_indices]

    if record.sleep_stages is None or sleep_stage_duration == record.sleep_stage_duration:
        stages = record.sleep_stages
    else:
        stages = interp1d(
            np.arange(len(record.sleep_stages)) * record.sleep_stage_duration,
            record.sleep_stages,
            kind='nearest',
            bounds_error=False,
            fill_value=(record.sleep_stages[0], record.sleep_stages[-1]),
        )(stage_times)

    return features, stages


def extract_features(
    records: Iterable[SleepRecord],
    lookback: int = 0,
    lookforward: int = 30,
    sleep_stage_duration: int = 30,
    feature_selection: Optional[List[str]] = None,
    fs_rri_resample: float = 4,
    min_rri: Optional[float] = None,
    max_rri: Optional[float] = None,
    max_nans: float = 0,
    n_jobs: int = 1,
) -> Tuple[List[np.ndarray], List[Union[np.ndarray, None]], List[str]]:
    """
    Calculate features from sleep data (e.g. heart rate).

    Time and frequency domain heart rate variability (HRV) features are
    calculated based on [1]_, [2]_ and [3]_. :ref:`feature_extraction`
    lists all available features and feature groups.

    Parameters
    ----------
    records : Iterable[SleepRecord]
        An iterable of `SleepRecord` objects, as yielded by the various
        reader functions in SleepECG.
    lookback : int, optional
        Backward extension of the analysis window from each sleep stage
        time in seconds, by default `0`.
    lookforward : int, optional
        Forward extension of the analysis window from each sleep stage
        time in seconds, by default `30`.
    sleep_stage_duration : int, optional
        Duration of a single sleep stage in the returned `stages` in
        seconds, by default `30`.
    feature_selection : list[str], optional
        Which features to extract. Can be feature groups or single feature
        identifiers, as listed :ref:`here<feature_extraction>`. If
        `None` (default), all possible features are extracted.
    fs_rri_resample : float, optional
        Frequency in Hz at which the RRI time series should be resampled
        before spectral analysis. Only relevant for frequency domain
        features, by default `4`.
    min_rri: float, optional
        Minimum RRI value in seconds to be considered valid. Will be passed
        to :func:`preprocess_rri`, by default `None`.
    max_rri: float, optional
        Maximum RRI value in seconds to be considered valid. Will be passed
        to :func:`preprocess_rri`, by default `None`.
    max_nans : float, optional
        Maximum fraction of NaNs in an analysis window, for which frequency
        features are computed. Should be a value between `0.0` and `1.0`,
        by default `0`.
    n_jobs : int, optional
        The number of jobs to run in parallel. If `1` (default), no
        parallelism is used. `-1` means using all processors.

    Returns
    -------
    features : list[np.ndarray]
        A list containing feature matrices which are arrays of shape
        `(len(sleep_stages), <num_features>)` and contain the extracted
        features per record.
    stages : list[np.ndarray | None]
        A list containing label vectors, i.e. the annotated sleep stages.
        For any `SleepRecord` without annotated stages, the corresponding
        list entry will be `None`.
    feature_ids : list[str]
        A list containing the identifiers of the extracted features.
        Feature groups passed in `feature_selection` are expanded to all
        individual features they contain. The order matches the column
        order of `X`.

    Notes
    -----
    .. [1] Task Force of the European Society of Cardiology. (1996). Heart
       rate variability: standards of measurement, physiological
       interpretation and clinical use. circulation, 93, 1043-1065.
       https://doi.org/10.1161/01.CIR.93.5.1043
    .. [2] Shaffer, F., & Ginsberg, J. P. (2017). An overview of heart rate
       variability metrics and norms. Frontiers in public health, 258.
       https://doi.org/10.3389/fpubh.2017.00258
    .. [3] Toichi, M., Sugiura, T., Murai, T., & Sengoku, A. (1997). A new
       method of assessing cardiac autonomic function and its comparison
       with spectral analysis and coefficient of variation of R–R interval.
       Journal of the autonomic nervous system, 62(1-2), 79-84.
       https://doi.org/10.1016/S0165-1838(96)00112-9
    """
    if feature_selection is None:
        feature_selection = list(_FEATURE_GROUPS)

    required_groups, feature_ids, col_indices = _parse_feature_selection(feature_selection)

    # _extract_features_single has two return values, so the list returned
    # by _parallel needs to be unpacked
    Xy = _parallel(
        n_jobs,
        _extract_features_single,
        records,
        sleep_stage_duration,
        min_rri,
        max_rri,
        required_groups,
        lookback,
        lookforward,
        fs_rri_resample,
        max_nans,
        feature_ids,
        col_indices,
    )
    features = [X for X, _ in Xy]
    stages = [y for _, y in Xy]

    return features, stages, feature_ids



# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""Heartbeat detection and detector evaluation."""

import warnings
from typing import NamedTuple

import numpy as np
import scipy.interpolate
import scipy.signal
import scipy.stats

_all_backends = ('c', 'numba', 'python')
_available_backends = list(_all_backends)

try:
    from ._heartbeat_detection import _squared_moving_integration, _thresholding
except ImportError:
    _available_backends.remove('c')

try:
    from numba import jit
except ImportError:
    _available_backends.remove('numba')


# cache sos-filter created with scipy.signal.butter to reduce runtime
_sos_filters = {}


def detect_heartbeats(ecg: np.ndarray, fs: float, backend: str = 'c') -> np.ndarray:
    """
    Detect heartbeats in an ECG signal.

    This is a modified version of the beat detection algorithm using
    adaptive thresholds described by Pan & Tompkins in 1985.

    Modifications/additions to the original algorithm described in
    https://doi.org/10.1109/TBME.1985.325532 are listed here:

    - Instead of a hardware filter adjusted to the sampling frequency of
      the MIT-BIH Arrhythmia Database, a 2nd-order bandpass with cutoff
      frequencies 5 and 30 Hz created via `scipy.signal.butter` is used.
    - A bidirectional filter is used to remove filter delay.
    - The signal might start during a peak, in which case it might have a
      relatively high initial amplitude, which will mess up threshold
      initialization. Therefore, everything until the first zero-crossing
      is set to zero.
    - The integration window is centered on the filtered signal, i.e. a
      peak in the filtered signal corresponds to a plateau in the
      integrated one, not a saddle in the rising edge. This lets the
      adaptive threshold for the integrated signal remain at a higher
      level, which is less susceptible to noise.
    - Learning phase 1 is not described in detail in the original paper.
      This implementation uses maximum and mean values inside the first 2
      seconds to initialize SPKI/SPKF/NPKI/NPKF. Details are provided in
      the `_thresholding` code.
    - In addition to the original searchback criterion, a searchback is
      also performed if no peak is found during the first second of the
      signal or no second peak is found 1.5s after the first one. This
      ensures correct behaviour at signal start in case an unusually large
      peak during learning phase 1 messes up threshold initialization.
    - After an unsuccessful searchback, the procedure is repeated in the
      same interval with further reduced thresholds, up to 16 times.

    Parameters
    ----------
    ecg : np.ndarray
        ECG signal.
    fs : float
        Sampling frequency in Hz.
    backend : {'c', 'numba', 'python'}
        Which implementation of the squared moving integration and
        thresholding algorithm to use. If available, `'c'` is the fastest
        implementation, `'numba'` is about 25% slower, and `'python'` is
        about 20 times slower but provided as a fallback. By default `'c'`.

    Returns
    -------
    np.ndarray
        Indices of detected heartbeats.
    """
    if backend not in _all_backends:
        raise ValueError(
            f'Invalid backend for detect_heartbeats: {backend!r}. '
            f'Possible options are: {_all_backends}.',
        )
    if backend not in _available_backends:
        fallback = _available_backends[0]
        warnings.warn(f'Backend {backend!r} not available, using {fallback!r} instead.')
        backend = fallback

    # For short signals, creating the bandpass filter makes up a large part
    # of the total runtime. Therefore the filter is cached in a global
    # variable.
    try:
        sos = _sos_filters[fs]
    except KeyError:
        sos = scipy.signal.butter(
            N=2,
            Wn=(5, 30),
            btype='bandpass',
            output='sos',
            fs=fs,
        )
        _sos_filters[fs] = sos

    # filtering bidirectionally removes filter delay
    filtered_ecg = scipy.signal.sosfiltfilt(sos, ecg)

    # Set everything until the first zero-crossing to zero. For efficiency,
    # only the first 2 seconds are checked.
    signal_start = np.where(np.diff(np.signbit(filtered_ecg[:int(2 * fs)])))[0][0] + 1
    filtered_ecg[:signal_start] = 0

    # scipy.signal.sosfiltfilt returns an array with negative strides. Both
    # `np.correlate` and `_thresholding` require contiguity, so ensuring
    # this here once reduces total runtime.
    filtered_ecg = np.ascontiguousarray(filtered_ecg)

    # five-point derivative as described by Pan & Tompkins
    derivative = np.correlate(filtered_ecg, np.array([-1, -2, 0, 2, 1]), mode='same')

    moving_window_width = int(0.15 * fs)  # 150ms

    if backend == 'c':
        integrated_ecg = _squared_moving_integration(derivative, moving_window_width)
        beat_mask = _thresholding(filtered_ecg, integrated_ecg, fs)
    elif backend == 'numba':
        integrated_ecg = _squared_moving_integration_numba(derivative, moving_window_width)
        beat_mask = _thresholding_numba(filtered_ecg, integrated_ecg, fs)
    elif backend == 'python':
        integrated_ecg = np.convolve(
            derivative**2,
            np.ones(moving_window_width),
            mode='same',
        )
        beat_mask = _thresholding_py(filtered_ecg, integrated_ecg, fs)

    return np.where(beat_mask)[0]


class _CompareHeartbeatsResult(NamedTuple):
    TP: np.ndarray
    FP: np.ndarray
    FN: np.ndarray


def compare_heartbeats(
    detection: np.ndarray,
    annotation: np.ndarray,
    max_distance: int = 0,
) -> _CompareHeartbeatsResult:
    """
    Determine correctness of detection results.

    Determine true positives (TP), false positives (FP) and false negatives
    (FN) for an array of detected heartbeat indices based on an array of
    annotated heartbeat indices. Since neither annotations nor automated
    detectors usually hit the peak perfectly, detected peaks no further
    than `max_distance` in both directions from an annotated peak are
    considered true positives.

    Parameters
    ----------
    detection : np.ndarray
        Detected heartbeat indices.
    annotation : np.ndarray
        Annotated heartbeat indices.
    max_distance : int, optional
        Maximum distance between indices to consider as the same peak, by
        default `0`.

    Returns
    -------
    TP : np.ndarray
        True positives, i.e. actual heartbeats detected as heartbeats.
    FP : np.ndarray
        False positives, i.e. non-heartbeats detected as heartbeats.
    FN : np.ndarray
        False negatives, i.e. actual heartbeats not detected as heartbeats.
    """
    if len(detection) == 0:
        return _CompareHeartbeatsResult(
            TP=np.array([]),
            FP=np.array([]),
            FN=annotation.copy(),
        )

    max_len = max(np.max(detection), np.max(annotation)) + 1
    detection_mask = np.zeros(max_len, dtype=bool)
    detection_mask[detection] = 1
    annotation_mask = np.zeros(max_len, dtype=bool)
    annotation_mask[annotation] = 1

    fuzzy_filter = np.ones(max_distance * 2 + 1, dtype=bool)
    detection_mask_fuzzy = np.convolve(detection_mask, fuzzy_filter, mode='same')
    annotation_mask_fuzzy = np.convolve(annotation_mask, fuzzy_filter, mode='same')

    return _CompareHeartbeatsResult(
        TP=np.where(detection_mask & annotation_mask_fuzzy)[0],
        FP=np.where(detection_mask & ~annotation_mask_fuzzy)[0],
        FN=np.where(annotation_mask & ~detection_mask_fuzzy)[0],
    )


class _RRISimilarityResult(NamedTuple):
    pearsonr: float
    spearmanr: float
    rmse: float


def rri_similarity(
    detection: np.ndarray,
    annotation: np.ndarray,
    fs_resample: float = 4,
) -> _RRISimilarityResult:
    """
    Calculate measures of similarity between RR intervals.

    RR intervals are calculated from detected and annotated heartbeat
    indices. The RR time series is then resampled to frequency
    `fs_resample` in the timespan common to both detection and annotation.
    Pearson's and Spearman's correlation coefficients as well as the root
    mean square error are returned.

    Parameters
    ----------
    detection : np.ndarray
        Detected heartbeat indices.
    annotation : np.ndarray
        Annotated heartbeat indices.

    Returns
    -------
    pearsonr : float
        Pearson correlation coefficient between resampled RR time series.
    spearmanr : float
        Spearman correlation coefficient between resampled RR time series.
    rmse : float
        Root mean square error between resampled RR time series.
    """
    rr_ann = np.diff(annotation)
    rr_det = np.diff(detection)

    t_ann = annotation[1:]
    t_det = detection[1:]

    start = max(min(t_ann), min(t_det))
    end = min(max(t_ann), max(t_det))
    t_new = np.arange(start, end, 1/fs_resample)

    interp_det = scipy.interpolate.interp1d(t_det, rr_det)
    interp_ann = scipy.interpolate.interp1d(t_ann, rr_ann)
    det_resampled = interp_det(t_new)
    ann_resampled = interp_ann(t_new)

    return _RRISimilarityResult(
        scipy.stats.pearsonr(det_resampled, ann_resampled)[0],
        scipy.stats.spearmanr(det_resampled, ann_resampled)[0],
        np.sqrt(np.mean((det_resampled - ann_resampled)**2)),
    )


def _squared_moving_integration_py(x: np.ndarray, window_length: int) -> np.ndarray:
    """
    Perform squaring and moving integration of an array.

    Parameters
    ----------
    x : np.ndarray
        The input array.
    window_length : int
        Length of the moving window in samples.

    Returns
    -------
    np.ndarray
        The squared and integrated array.
    """
    signal_len = len(x)
    if not 0 < window_length <= signal_len:
        raise ValueError('window_length has to be 0 < window_length <= len(x)')

    output = np.empty_like(x)

    # create a circular buffer to store values inside integration window
    integration_buffer = np.zeros(window_length)
    sum = 0

    # the integration window is centered on the original signal, for even
    # window_length the behavior of np.convolve with a constant window of
    # even length is replicated (i.e the window is off-center to the left)
    window_length_half = (window_length + 1) // 2

    # during the first `window_length/2` samples there is no output since
    # the integration window's center would be at a negative index of the
    # input
    for i in range(window_length_half):
        square = x[i] * x[i]
        integration_buffer[i % window_length] = square
        sum += square

    for i in range(window_length_half, signal_len):
        output[i - window_length_half] = sum  # write to 'window center'
        sum -= integration_buffer[i % window_length]
        square = x[i] * x[i]
        integration_buffer[i % window_length] = square
        sum += square

    # the end of the x signal is reached, so the integration window is
    # built down and the last `window_length/2` entries of the output are
    # filled
    for i in range(signal_len, signal_len+window_length_half):
        output[i - window_length_half] = sum
        sum -= integration_buffer[i % window_length]

    return output


def _thresholding_py(
    filtered_ecg: np.ndarray,
    integrated_ecg: np.ndarray,
    fs: float,
) -> np.ndarray:
    """
    Perform adaptive thresholding based on Pan and Tompkin's algorithm.

    Parameters
    ----------
    filtered_ecg : np.ndarray
        Bandpass filtered ECG.
    integrated_ecg : np.ndarray
        Squared and moving window integrated bandpass filtered ECG.
    fs : float
        Sampling frequency in Hz.

    Returns
    -------
    np.ndarray
        Array containing `1` for every sample in `filtered_ecg` identified
        as a heartbeat.
    """
    signal_len = len(filtered_ecg)
    beat_mask = np.zeros_like(filtered_ecg)  # numba can't work with dtype=bool

    REFRACTORY_SAMPLES = int(0.2 * fs)  # 200ms
    T_WAVE_WINDOW = int(0.36 * fs)  # 360ms

    # --------------------------------------------------------------------
    # Learning Phase 1
    # --------------------------------------------------------------------
    # Pan & Tompkins mention a learning phase to initialize detection
    # thresholds based upon signal and noise peaks detected during the
    # first two seconds. The exact initialization process is not
    # described. The adaptive thresholds are calculated based on running
    # estimates of signal and noise peaks (`SPKF` and `NPKF` for the
    # filtered signal). Assuming constant peak amplitudes, those values
    # converge towards the signal peak amplitude and noise peak amplitude,
    # respectively. Therefore, SPKF/SPKI are assumed to be the maximum
    # values of the filtered/integrated signal during the learning phase.
    # Accordingly, NPKF/NPKI are initialized to the mean values during the
    # first two seconds.
    SPKF = np.max(filtered_ecg[:int(2 * fs)])
    NPKF = np.mean(filtered_ecg[:int(2 * fs)])
    SPKI = np.max(integrated_ecg[:int(2 * fs)])
    NPKI = np.mean(integrated_ecg[:int(2 * fs)])
    threshold_I1 = NPKI + 0.25 * (SPKI - NPKI)
    threshold_F1 = NPKF + 0.25 * (SPKF - NPKF)

    # According to the original paper, `RR AVERAGE2` is the average of the
    # last 8 RR intervals that lie in a certain interval. In the worst
    # case, this requires going back to the very first RR interval.
    # Therefore, all RR intervals are stored. As the algorithm enforces a
    # refractory period, the maximum number of heartbeats is equal to
    # signal_len / refractory_samples.
    RR_intervals = np.zeros(signal_len//REFRACTORY_SAMPLES)

    # tracking the number of peaks found is required to calculate the
    # average RR intervals correctly during the first 8 beats (also needed
    # for access to RR_intervals)
    num_peaks_found = 0

    RR_missed_limit = None

    # in case a searchback was unsuccessful, no new searchback will be
    # performed until another signal peak has been found regularly
    do_searchback = True

    # initialize, so searchback before any peak has been detected works
    peak_index = -REFRACTORY_SAMPLES + 1
    previous_peak_index = -REFRACTORY_SAMPLES + 1

    index = 1
    while index < signal_len - 1:

        PEAKF = None
        PEAKI = None

        signal_peak_found = False
        noise_peak_found = False
        # ----------------------------------------------------------------
        # Searchback
        # ----------------------------------------------------------------
        # During a "searchback", detection thresholds are reduced by one
        # half. The peak with highest amplitude between 200ms (i.e. the
        # refractory period) after the previous detected peak and the
        # current index is considered as a peak candidate.
        # Modifications compared to Pan & Tompkins' original method:
        # - The original paper states that a searchback peak's amplitude
        #   has to be between the original threshold and the reduced one.
        #   It can happen that this is the case for the filtered signal,
        #   but not for the integrated one (if the raw signal amplitude is
        #   suddenly considerably lower). Therefore, this implementation
        #   requires both signals (filtered and integrated) to be above
        #   the reduced threshold, but only one of them to be below the
        #   original threshold.
        # - No further steps are specified for the case that no peak is
        #   found during searchback. Since a searchback is triggered
        #   because (physiologically) there has to be a heartbeat during
        #   the searchback interval, this implementation repeats the
        #   process with further reduced thresholds. Up to 16 searchback
        #   runs are performed, each time the thresholds are further
        #   reduced by 1/2. A hard limit of 16 runs avoids an endless loop
        #   in case there's really just noise.
        # - Since the criterion for triggering a searchback is based on
        #   the average RR interval, in the original form this could only
        #   happen after at least two detected heartbeats. An
        #   exceptionally large peak during the first learning phase can
        #   throw the initial thresholds off, so peaks at the beginning
        #   are ignored - which in turn invalidates learning phase 2.
        #   Therefore, in addition to the original searchback criterion
        #   (no peak during 1.66 * "the average RR interval"), a
        #   searchback is triggered in two cases: (1) if there is no peak
        #   during the first second, and (2) if there is no peak 1.5s
        #   after the first peak.
        if ((num_peaks_found > 1 and index - previous_peak_index > RR_missed_limit and do_searchback) or  # original criterion  # noqa: E501
                (num_peaks_found == 0 and index > fs) or                             # (1)
                (num_peaks_found == 1 and index - previous_peak_index > 1.5 * fs)):  # (2)

            for i in range(1, 16):
                found_a_candidate = False

                searchback_divisor = 1 << i  # 2^i
                best_searchback_index = previous_peak_index + REFRACTORY_SAMPLES
                best_candidate_amplitude = -1
                searchback_index = best_searchback_index

                while searchback_index < index:
                    PEAKF = filtered_ecg[searchback_index]
                    if PEAKF > filtered_ecg[searchback_index + 1]:
                        # next one is lower
                        if PEAKF > filtered_ecg[searchback_index - 1]:
                            # it is a peak
                            PEAKI = integrated_ecg[searchback_index]
                            # one signal is between the reduced and
                            # original threshold, the other one above the
                            # reduced threshold
                            if ((threshold_F1 / searchback_divisor < PEAKF and PEAKF < threshold_F1 and threshold_I1 / searchback_divisor < PEAKI) or  # noqa: E501
                                    (threshold_I1 / searchback_divisor < PEAKI and PEAKI < threshold_I1 and threshold_F1 / searchback_divisor < PEAKF)):   # noqa: E501
                                if PEAKF > best_candidate_amplitude:
                                    # highest one so far
                                    best_searchback_index = searchback_index
                                    best_candidate_amplitude = filtered_ecg[searchback_index]  # noqa: E501
                                    found_a_candidate = True

                        # the amplitude of the next sample is lower, so it
                        # can't be a peak -> skip it
                        searchback_index += 1

                    searchback_index += 1

                if found_a_candidate:
                    SPKI = 0.25 * PEAKI + 0.75 * SPKI
                    SPKF = 0.25 * PEAKF + 0.75 * SPKF
                    signal_peak_found = True
                    peak_index = best_searchback_index

                    # don't perform a searchback until the next signal
                    # peak has been found to avoid endless loops
                    do_searchback = False
                    break

        elif filtered_ecg[index] > filtered_ecg[index + 1]:
            if filtered_ecg[index] > filtered_ecg[index - 1]:
                # a local maximum in the filtered signal was found
                PEAKF = filtered_ecg[index]
                PEAKI = integrated_ecg[index]
                if PEAKF > threshold_F1 and PEAKI > threshold_I1:
                    # Both the filtered and the integrated signal are
                    # above their respective thresholds. Thus the current
                    # peak is classified as a signal peak and the running
                    # estimates SPKF and SPKI are updated.
                    SPKF = 0.125 * PEAKF + 0.875 * SPKF
                    SPKI = 0.125 * PEAKI + 0.875 * SPKI

                    signal_peak_found = True
                    peak_index = index
                else:
                    noise_peak_found = True

            # The next sample's amplitude is lower, meaning it can't be a
            # peak, so we skip it. This is why there are two separate
            # if-clauses for this block.
            index += 1

        # Calculating the RR interval and comparing slopes only makes
        # sense if there has already been a signal peak in the past.
        if signal_peak_found and num_peaks_found > 0:
            RR = peak_index - previous_peak_index

            # ------------------------------------------------------------
            # T Wave Identification
            # ------------------------------------------------------------
            # "When an RR interval is less than 360 ms (it must be greater
            # than the 200 ms latency), a judgment is made to determine
            # whether the current QRS complex has been correctly
            # identified or whether it is really a T wave. If the maximal
            # slope that occurs during this waveform is less than half
            # that of the QRS waveform that preceded it, it is identified
            # to be a T wave; otherwise, it is called a QRS complex."
            # (from Pan & Tompkins, 1985)
            if RR < T_WAVE_WINDOW:
                reverse_index = peak_index
                max_slope_in_this_peak = -1
                while reverse_index > 0:
                    amplitude_here = filtered_ecg[reverse_index]
                    amplitude_before = filtered_ecg[reverse_index - 1]
                    if amplitude_before > amplitude_here:
                        break
                    slope = amplitude_here - amplitude_before
                    if slope > max_slope_in_this_peak:
                        max_slope_in_this_peak = slope
                    reverse_index -= 1

                reverse_index = previous_peak_index
                max_slope_in_previous_peak = -1
                while reverse_index > 0:
                    amplitude_here = filtered_ecg[reverse_index]
                    amplitude_before = filtered_ecg[reverse_index - 1]
                    if amplitude_before > amplitude_here:
                        break
                    slope = amplitude_here - amplitude_before
                    if (slope > max_slope_in_previous_peak):
                        max_slope_in_previous_peak = slope
                    reverse_index -= 1

                if max_slope_in_this_peak < max_slope_in_previous_peak / 2.0:
                    # based on the slope, this peak should be a T wave
                    signal_peak_found = False
                    noise_peak_found = True

        if signal_peak_found:
            # What we know so far: we are at a local maximum, both
            # thresholds are exceeded and it is not a T wave. Thus, the
            # current sample can be considered as a "signal peak" and the
            # adaptive thresholds are updated.
            num_peaks_found += 1
            beat_mask[peak_index] = 1

            threshold_I1 = NPKI + 0.25 * (SPKI - NPKI)
            threshold_F1 = NPKF + 0.25 * (SPKF - NPKF)

            # calculating RR averages only makes sense once 2 peaks have
            # been found
            if num_peaks_found > 1:
                RR_intervals[num_peaks_found] = peak_index - previous_peak_index

                # --------------------------------------------------------
                # Learning phase 2
                # --------------------------------------------------------
                # "Learning phase 2 requires two heartbeats to initialize
                # RR interval average and RR interval limit values."
                # (from Pan & Tompkins, 1985)
                if num_peaks_found == 2:
                    RR_low_limit = 0.92 * RR_intervals[num_peaks_found]
                    RR_high_limit = 1.16 * RR_intervals[num_peaks_found]

                # --------------------------------------------------------
                # RR Average 1 / RR Average 2
                # --------------------------------------------------------
                # RR Average 2 is the average of the 8 most recent RR
                # intervals which fell between RR_low_limit and
                # RR_high_limit. In case of a regular heart rate, this
                # equals RR Average 1 (the average over the 8 most recent
                # RR intervals, independent of any limits). Therefore, RR
                # Average 1 does not need to be calculated separately.
                RR_sum = 0
                RR_count = 0
                irregular = False
                for i in range(num_peaks_found, 1, -1):
                    RR_n = RR_intervals[i]
                    if RR_low_limit < RR_n and RR_n < RR_high_limit:
                        RR_sum += RR_n
                        RR_count += 1
                        if RR_count >= 8:
                            break
                    else:
                        irregular = True

                RR_average = RR_sum / RR_count
                RR_low_limit = 0.92 * RR_average
                RR_high_limit = 1.16 * RR_average
                RR_missed_limit = 1.66 * RR_average

                if irregular:
                    # "For irregular heart rates, the first threshold of
                    # each set is reduced by half so as to increase the
                    # detection sensitivity and to avoid missing beats."
                    threshold_F1 /= 2
                    threshold_I1 /= 2

            # A signal peak has been found, so performing a searchback
            # makes sense.
            do_searchback = True

            # previous peak index is required to calculate the RR interval
            previous_peak_index = peak_index

            # no peak can happen during the refractory period, so skip it
            index = peak_index + REFRACTORY_SAMPLES

        elif noise_peak_found:
            NPKI = 0.125 * PEAKI + 0.875 * NPKI
            NPKF = 0.125 * PEAKF + 0.875 * NPKF
            threshold_I1 = NPKI + 0.25 * (SPKI - NPKI)
            threshold_F1 = NPKF + 0.25 * (SPKF - NPKF)

        index += 1

    # return an array containing `1`s at beat positions and `0`s elsewhere
    return beat_mask


if 'numba' in _available_backends:
    _squared_moving_integration_numba = jit(_squared_moving_integration_py)
    _thresholding_numba = jit(_thresholding_py)



# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""Utility functions."""


import warnings
from typing import Callable, Iterable, List, TypeVar

# required to propagate the return type annotation through _parallel
_Returnable = TypeVar('_Returnable')


def _parallel(
    n_jobs: int,
    function: Callable[..., _Returnable],
    iterable: Iterable,
    *args,
    **kwargs,
) -> List[_Returnable]:
    """
    Apply a function to each element in an iterable in parallel.

    This uses joblib for parallelism. If the package is not available, it
    falls back to a pure Python loop.

    Parameters
    ----------
    n_jobs : int
        The number of jobs to run in parallel. If `1` (default), no
        parallelism is used. `-1` means using all processors.
    function : Callable
        The function to call.
    iterable : Iterable
        `function` will be applied to every element in this iterable.
    *args
        Positional arguments to be passed to `function`.
    **kwargs
        Keyword arguments to be passed to `function`.

    Returns
    -------
    list
        A list containing the return values of `function`.

    Warnings
    --------
    Note that in case the `function` is very simple, the cost for spawning
    workers will make the parallel execution slower than the standard
    execution.

    Examples
    --------
    To parallelize this:

    >>> def root(a, n):
    >>>     return a**(1/n)
    >>> res = [root(x, 3) for x in range(1000)]

    Write this:

    >>> n_jobs = 4
    >>> res = _parallel(n_jobs, root, range(1000), 3)
    """
    try:
        from joblib import Parallel, delayed
    except ImportError:
        if n_jobs != 1:
            warnings.warn('joblib not installed, cannot run in parallel.')
        return [function(x, *args, **kwargs) for x in iterable]

    return Parallel(n_jobs=n_jobs)(
        delayed(function)(x, *args, **kwargs) for x in iterable
    )



# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""Interface for listing and downloading NSRR (sleepdata.org) data."""

from fnmatch import fnmatch
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import List, Tuple, Union

import requests
from tqdm import tqdm

from .utils import _download_file

_nsrr_token = None


def set_nsrr_token(token: str) -> None:
    """
    Set and verify the NSRR (sleepdata.org) download token.

    Implemented according to the NSRR API documentation:
    https://github.com/nsrr/sleepdata.org/wiki/api-v1-account

    Parameters
    ----------
    token : str
        NSRR (sleepdata.org) download token (get it from
        https://sleepdata.org/token).

    """
    response = requests.get(
        'https://sleepdata.org/api/v1/account/profile.json',
        params={'auth_token': token},
    )
    authenticated = response.json()['authenticated']
    if authenticated:
        username = response.json()['username']
        email = response.json()['email']
        print(f'Authenticated at sleepdata.org as {username} ({email})')
        global _nsrr_token
        _nsrr_token = token
    else:
        raise RuntimeError('Authentication at sleepdata.org failed, verify token!')


def _get_nsrr_url(db_slug: str) -> str:
    """
    Get the download URL for a given NSRR database.

    The download token is a part of the URL, so it needs to be already set.

    Parameters
    ----------
    db_slug : str
        Short identifier of a database, e.g. `'mesa'`.

    Returns
    -------
    str
        The download URL.
    """
    if _nsrr_token is None:
        raise RuntimeError('NSRR token not set, use `sleepecg.set_nsrr_token(<token>)`!')
    return f'https://sleepdata.org/datasets/{db_slug}/files/a/{_nsrr_token}/m/sleepecg/'


def _list_nsrr(
    db_slug: str,
    subfolder: str = '',
    pattern: str = '*',
    shallow: bool = False,
) -> List[Tuple[str, str]]:
    """
    Recursively list filenames and checksums for a dataset.

    Specify a subfolder and/or a filename-pattern to filter results.

    Implemented according to the NSRR API documentation:
    https://github.com/nsrr/sleepdata.org/wiki/api-v1-datasets#list-files-in-folder

    Parameters
    ----------
    db_slug : str
        Short identifier of a database, e.g. `'mesa'`.
    subfolder : str, optional
        The folder at which to start the search, by default `''` (i.e. the
        root folder).
    pattern : str, optional
        Glob-like pattern to select files (only applied to the basename,
        not the dirname), by default `'*'`.
    shallow : bool, optional
        If `True`, only search in the given subfolder (i.e. no recursion),
        by default `False`.

    Returns
    -------
    list[tuple[str, str]]
        A list of tuples `(<filename>, <checksum>)`; `<filename>` is the
        full filename (i.e. dirname and basename) and `<checksum>` the
        MD5 checksum.
    """
    api_url = f'https://sleepdata.org/api/v1/datasets/{db_slug}/files.json'

    response = requests.get(api_url, params={'path': subfolder})
    try:
        response_json = response.json()
    except JSONDecodeError:
        raise RuntimeError(f'No API response for dataset {db_slug}.') from None

    files = []
    for item in response_json:
        if not item['is_file'] and not shallow:
            files.extend(_list_nsrr(db_slug, item['full_path'], pattern))
        elif fnmatch(item['file_name'], pattern):
            files.append((item['full_path'], item['file_checksum_md5']))
    return files


def _download_nsrr_file(
    url: str,
    target_filepath: Path,
    checksum: str,
):
    """
    Download a file from `url` to `target_filepath` and verify `checksum`.

    This is a wrapper around `sleepecg.io.utils._download_file` to provide
    a helpful error message in case the currently set token does not grant
    access to the requested file.

    Parameters
    ----------
    url : str
        URL to download from.
    target_filepath : pathlib.Path
        Location where the downloaded file will be stored.
    checksum : str
        Checksum to verify the file against.
    """
    try:
        _download_file(url, target_filepath, checksum, 'md5')
    except RuntimeError as error:
        # If the token is invalid for the requested dataset, the
        # request is redirected to a files-overview page. The response
        # is an HTML-page which doesn't have a "content-disposition"
        # header.
        response = requests.get(url, stream=True)
        if 'content-disposition' not in response.headers:
            db_slug = url.split('/')[4]
            raise RuntimeError(f'Make sure you have access to {db_slug}!') from error
        else:
            raise


def download_nsrr(
    db_slug: str,
    subfolder: str = '',
    pattern: str = '*',
    shallow: bool = False,
    data_dir: Union[str, Path] = '.',
) -> None:
    """
    Recursively download files from NSRR (sleepdata.org).

    Specify a subfolder and/or a filename-pattern to filter results.

    Implemented according to the NSRR API documentation:
    https://github.com/nsrr/sleepdata.org/wiki/api-v1-datasets#download-a-file

    Parameters
    ----------
    db_slug : str
        Short identifier of a database, e.g. `'mesa'`.
    subfolder : str, optional
        The folder at which to start the search, by default `''` (i.e. the
        root folder).
    pattern : str, optional
        Glob-like pattern to select files (only applied to the basename,
        not the dirname), by default `'*'`.
    shallow : bool, optional
        If `True`, only download files in the given subfolder (i.e. no
        recursion), by default `False`.
    data_dir : str | pathlib.Path, optional
        Directory where all datasets are stored, by default `'.'`.
    """
    db_dir = Path(data_dir) / db_slug

    download_url = _get_nsrr_url(db_slug)
    files_to_download = _list_nsrr(db_slug, subfolder, pattern, shallow)
    tqdm_description = f'Downloading {db_slug}/{subfolder or "."}/{pattern}'

    for filepath, checksum in tqdm(files_to_download, desc=tqdm_description):
        target_filepath = db_dir / filepath
        url = download_url + filepath
        _download_nsrr_file(url, target_filepath, checksum)



# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""Read datasets containing ECG data and sleep stage annotations."""

import datetime
from dataclasses import dataclass
from enum import IntEnum
from pathlib import Path
from typing import Iterator, NamedTuple, Optional, Union
from xml.etree import ElementTree

import numpy as np

from ..config import get_config
from ..heartbeats import detect_heartbeats
from .nsrr import _download_nsrr_file, _get_nsrr_url, _list_nsrr
from .physionet import _list_physionet, download_physionet


class SleepStage(IntEnum):
    """
    Mapping of AASM sleep stages to integers.

    To facilitate hypnogram plotting, values increase with wakefulness.
    """

    WAKE = 5
    REM = 4
    N1 = 3
    N2 = 2
    N3 = 1
    UNDEFINED = -1


class Gender(IntEnum):
    """Mapping of gender to integers."""

    FEMALE = 0
    MALE = 1


@dataclass
class SubjectData:
    """
    Store data about a single subject.

    Attributes
    ----------
    gender : int, optional
        The subject's gender, stored as an integer as defined by `Gender`,
        by default `None`.
    age : int, optional
        The subject's age in years, by default `None`.
    weight : int, optional
        The subject's weight in kg, by default `None`.
    """

    gender: Optional[int] = None
    age: Optional[int] = None
    weight: Optional[int] = None


@dataclass
class SleepRecord:
    """
    Store a single sleep record.

    Attributes
    ----------
    sleep_stages : np.ndarray, optional
        Sleep stages according to AASM guidelines, stored as integers as
        defined by :class:`SleepStage`, by default `None`.
    sleep_stage_duration : int, optional
        Duration of each sleep stage in seconds, by default `None`.
    id : str, optional
        The record's ID, by default `None`.
    recording_start_time : datetime.time, optional
        Time at which the recording was started, by default `None`.
    heartbeat_times : np.ndarray, optional
        Times of heartbeats relative to recording start in seconds, by
        default `None`.
    subject_data : SubjectData, optional
        Dataclass containing subject data, such as gender or age, by
        default `None`.
    """

    sleep_stages: Optional[np.ndarray] = None
    sleep_stage_duration: Optional[int] = None
    id: Optional[str] = None
    recording_start_time: Optional[datetime.time] = None
    heartbeat_times: Optional[np.ndarray] = None
    subject_data: Optional[SubjectData] = None


class _ParseNsrrXmlResult(NamedTuple):
    sleep_stages: np.ndarray
    sleep_stage_duration: int
    recording_start_time: datetime.time


def _parse_nsrr_xml(xml_filepath: Path) -> _ParseNsrrXmlResult:
    """
    Parse NSRR XML sleep stage annotation file.

    Parameters
    ----------
    xml_filepath : pathlib.Path
        Path of the annotation file to read.

    Returns
    -------
    sleep_stages : np.ndarray
        Sleep stages according to AASM guidelines, stored as integers as
        defined by :class:`SleepStage`.
    sleep_stage_duration : int
        Duration of each sleep stage in seconds.
    recording_start_time : datetime.time
        Time at which the recording was started.

    """
    STAGE_MAPPING = {
        'Wake|0': SleepStage.WAKE,
        'Stage 1 sleep|1': SleepStage.N1,
        'Stage 2 sleep|2': SleepStage.N2,
        'Stage 3 sleep|3': SleepStage.N3,
        'Stage 4 sleep|4': SleepStage.N3,
        'REM sleep|5': SleepStage.REM,
        'Unscored|9': SleepStage.UNDEFINED,
    }

    root = ElementTree.parse(xml_filepath).getroot()

    epoch_length = root.findtext('EpochLength')
    if epoch_length is None:
        raise RuntimeError(f'EpochLength not found in {xml_filepath}.')
    epoch_length = int(epoch_length)

    start_time = None
    annot_stages = []

    for event in root.find('ScoredEvents'):
        if event.find('EventConcept').text == 'Recording Start Time':
            start_time = event.find('ClockTime').text.split()[1]
            start_time = datetime.datetime.strptime(start_time, '%H.%M.%S').time()

        if event.find('EventType').text == 'Stages|Stages':
            epoch_duration = int(float(event.findtext('Duration')))
            stage = STAGE_MAPPING[event.findtext('EventConcept')]
            annot_stages.extend([stage] * int(epoch_duration / epoch_length))

    if start_time is None:
        raise RuntimeError(f'"Recording Start Time" not found in {xml_filepath}.')

    return _ParseNsrrXmlResult(
        np.array(annot_stages, dtype=np.int8),
        epoch_length,
        start_time,
    )


def read_mesa(
    records_pattern: str = '*',
    heartbeats_source: str = 'annotation',
    offline: bool = False,
    keep_edfs: bool = False,
    data_dir: Optional[Union[str, Path]] = None,
) -> Iterator[SleepRecord]:
    """
    Lazily read records from MESA (https://sleepdata.org/datasets/mesa).

    Each MESA record consists of an `.edf` file containing raw
    polysomnography data and an `.xml` file containing annotated events.
    Since the entire MESA dataset requires about 385 GB of disk space,
    `.edf` files can be deleted after heartbeat times have been extracted.
    Heartbeat times are cached in an `.npy` file in
    `<data_dir>/mesa/preprocessed/heartbeats`.

    Parameters
    ----------
    records_pattern : str, optional
         Glob-like pattern to select record IDs, by default `'*'`.
    heartbeats_source : {'annotation', 'cached', 'ecg'}, optional
        If `'annotation'` (default), get heartbeat times from
        `polysomnography/annotations-rpoints/<record_id>-rpoints.csv` (not
        available for all records). If `'ecg'`, use
        `sleepecg.detect_heartbeats` on the ECG contained in
        `polysomnography/edfs/<record_id>.csv` and cache the result to
        `preprocessed/heartbeats/<record_id>.npy`. If `'cached'`, get the
        cached heartbeats.
    offline : bool, optional
        If `True`, search for local files only instead of using the NSRR
        API, by default `False`.
    keep_edfs : bool, optional
        If `False`, remove `.edf` after heartbeat detection, by default
        `False`.
    data_dir : str | pathlib.Path, optional
        Directory where all datasets are stored. If `None` (default), the
        value will be taken from the configuration.

    Yields
    ------
    SleepRecord
        Each element in the generator is a :class:`SleepRecord`.
    """
    from mne.io import read_raw_edf

    DB_SLUG = 'mesa'
    ANNOTATION_DIRNAME = 'polysomnography/annotations-events-nsrr'
    EDF_DIRNAME = 'polysomnography/edfs'
    HEARTBEATS_DIRNAME = 'preprocessed/heartbeats'
    RPOINTS_DIRNAME = 'polysomnography/annotations-rpoints'

    GENDER_MAPPING = {0: Gender.FEMALE, 1: Gender.MALE}

    heartbeats_source_options = {'annotation', 'cached', 'ecg'}
    if heartbeats_source not in heartbeats_source_options:
        raise ValueError(
            f'Invalid value for parameter `heartbeats_source`: {heartbeats_source}, '
            f'possible options: {heartbeats_source_options}',
        )

    if data_dir is None:
        data_dir = get_config('data_dir')

    db_dir = Path(data_dir).expanduser() / DB_SLUG
    annotations_dir = db_dir / ANNOTATION_DIRNAME
    edf_dir = db_dir / EDF_DIRNAME
    heartbeats_dir = db_dir / HEARTBEATS_DIRNAME

    for directory in (annotations_dir, edf_dir, heartbeats_dir):
        directory.mkdir(parents=True, exist_ok=True)

    if not offline:
        download_url = _get_nsrr_url(DB_SLUG)

        subject_data_filename, subject_data_checksum = _list_nsrr(
            'mesa',
            'datasets',
            'mesa-sleep-dataset-*.csv',
            shallow=True,
        )[0]
        subject_data_filepath = db_dir / subject_data_filename
        _download_nsrr_file(
            download_url + subject_data_filename,
            target_filepath=subject_data_filepath,
            checksum=subject_data_checksum,
        )

        checksums = {}
        xml_files = _list_nsrr(
            DB_SLUG,
            ANNOTATION_DIRNAME,
            f'mesa-sleep-{records_pattern}-nsrr.xml',
            shallow=True,
        )
        checksums.update(xml_files)
        requested_records = [Path(file).stem[:-5] for file, _ in xml_files]

        edf_files = _list_nsrr(
            DB_SLUG,
            EDF_DIRNAME,
            f'mesa-sleep-{records_pattern}.edf',
            shallow=True,
        )
        checksums.update(edf_files)

        rpoints_files = _list_nsrr(
            DB_SLUG,
            RPOINTS_DIRNAME,
            f'mesa-sleep-{records_pattern}-rpoint.csv',
            shallow=True,
        )
        checksums.update(rpoints_files)
    else:
        subject_data_filepath = next((db_dir / 'datasets').glob('mesa-sleep-dataset-*.csv'))
        xml_files = sorted(annotations_dir.glob(f'mesa-sleep-{records_pattern}-nsrr.xml'))
        requested_records = [file.stem[:-5] for file in xml_files]

    subject_data_array = np.loadtxt(
        subject_data_filepath,
        delimiter=',',
        skiprows=1,
        usecols=[0, 3, 5],  # [mesaid, gender, age]
        dtype=int,
    )

    subject_data = {}
    for mesaid, gender, age in subject_data_array:
        subject_data[f'mesa-sleep-{mesaid:04}'] = SubjectData(
            gender=GENDER_MAPPING[gender],
            age=age,
        )

    for record_id in requested_records:
        heartbeats_file = heartbeats_dir / f'{record_id}.npy'
        if heartbeats_source == 'annotation':
            rpoints_filename = f'{RPOINTS_DIRNAME}/{record_id}-rpoint.csv'
            rpoints_filepath = db_dir / rpoints_filename
            if not rpoints_filepath.is_file():
                if not offline and rpoints_filename in checksums:
                    _download_nsrr_file(
                        download_url + rpoints_filename,
                        rpoints_filepath,
                        checksums[rpoints_filename],
                    )
                else:
                    print(f'Skipping {record_id} due to missing heartbeat annotations.')
                    continue

            heartbeat_times = np.loadtxt(
                rpoints_filepath,
                delimiter=',',
                skiprows=1,
                usecols=18,  # column 18 ('seconds') contains the annotated heartbeat times
            )
            # for some reason some (39) records have unsorted annotations
            heartbeat_times.sort()
        elif heartbeats_source == 'cached':
            if not heartbeats_file.is_file():
                print(f'Skipping {record_id} due to missing cached heartbeats.')
                continue
            heartbeat_times = np.load(heartbeats_file)
        elif heartbeats_source == 'ecg':
            edf_filename = EDF_DIRNAME + f'/{record_id}.edf'
            edf_filepath = db_dir / edf_filename
            edf_was_available = edf_filepath.is_file()
            if not offline:
                _download_nsrr_file(
                    download_url + edf_filename,
                    edf_filepath,
                    checksums[edf_filename],
                )

            rec = read_raw_edf(edf_filepath, verbose=False)
            ecg = rec.get_data('EKG').ravel()
            fs = rec.info['sfreq']
            heartbeat_indices = detect_heartbeats(ecg, fs)
            heartbeat_times = heartbeat_indices / fs
            np.save(heartbeats_file, heartbeat_times)

            if not edf_was_available and not keep_edfs:
                edf_filepath.unlink()

        xml_filename = ANNOTATION_DIRNAME + f'/{record_id}-nsrr.xml'
        xml_filepath = db_dir / xml_filename
        if not offline:
            _download_nsrr_file(
                download_url + xml_filename,
                xml_filepath,
                checksums[xml_filename],
            )

        parsed_xml = _parse_nsrr_xml(xml_filepath)

        yield SleepRecord(
            sleep_stages=parsed_xml.sleep_stages,
            sleep_stage_duration=parsed_xml.sleep_stage_duration,
            id=record_id,
            recording_start_time=parsed_xml.recording_start_time,
            heartbeat_times=heartbeat_times,
            subject_data=subject_data[record_id],
        )


def read_slpdb(
    records_pattern: str = '*',
    offline: bool = False,
    data_dir: Optional[Union[str, Path]] = None,
) -> Iterator[SleepRecord]:
    """
    Lazily read records from SLPDB (https://physionet.org/content/slpdb).

    Required files are downloaded from PhysioNet to `<data_dir>/slpdb`.

    Parameters
    ----------
    records_pattern : str, optional
         Glob-like pattern to select record IDs, by default `'*'`.
    offline : bool, optional
        If `True`, search for local files only instead of downloading from
        PhysioNet, by default `False`.
    data_dir : str | pathlib.Path, optional
        Directory where all datasets are stored. If `None` (default), the
        value will be taken from the configuration.

    Yields
    ------
    SleepRecord
        Each element in the generator is a :class:`SleepRecord`.
    """
    # https://physionet.org/content/slpdb/1.0.0/
    import wfdb

    DB_SLUG = 'slpdb'

    STAGE_MAPPING = {
        'W': SleepStage.WAKE,
        'R': SleepStage.REM,
        '1': SleepStage.N1,
        '2': SleepStage.N2,
        '3': SleepStage.N3,
        '4': SleepStage.N3,
    }

    if data_dir is None:
        data_dir = get_config('data_dir')

    data_dir = Path(data_dir).expanduser()
    db_dir = data_dir / DB_SLUG

    requested_records = _list_physionet(
        data_dir=data_dir,
        db_slug=DB_SLUG,
        pattern=records_pattern,
    )

    if not offline:
        download_physionet(
            data_dir=data_dir,
            db_slug=DB_SLUG,
            requested_records=requested_records,
            extensions=['.hea', '.dat', '.st'],
        )

    for record_id in requested_records:
        record_file = str(db_dir / record_id)

        record = wfdb.rdrecord(record_file)
        start_time = record.base_time
        ecg = np.asarray(record.p_signal[:, record.sig_name.index('ECG')])
        fs = record.fs

        heartbeat_indices = detect_heartbeats(ecg, fs)
        heartbeat_times = heartbeat_indices / fs

        annot_st = wfdb.rdann(record_file, 'st')

        # Some 30 second windows don't have a sleep stage annotation, so
        # the annotation array is initialized with `SleepStage.UNDEFINED`
        # for every 30 second window.
        for sample_time, annotation in zip(annot_st.sample[::-1], annot_st.aux_note[::-1]):
            if annotation[0] in STAGE_MAPPING:
                number_of_sleep_stages = sample_time // (30 * fs) + 1
                break

        sleep_stages = np.full(number_of_sleep_stages, SleepStage.UNDEFINED)

        # Most annotations are at sample indices which are multiples of
        # 30*fs. However, annotations which would be at sample index 0, are
        # at sample index 1. Integer divison is used when calculating the
        # stage index to move these annotations to sample index 0.
        for sample_time, annotation in zip(annot_st.sample, annot_st.aux_note):
            if annotation[0] in STAGE_MAPPING:
                sleep_stages[sample_time // (30 * fs)] = STAGE_MAPPING[annotation[0]]

        # Age and weight are given in the last line of the header file,
        # which is contained in record.comments[0] and looks like this:
        # '44 M 89 32-01-89' ('<age> <gender> <weight> <unspecified>')
        # For some records, age/weight is given as 'x'.
        age, _, weight, _ = record.comments[0].split()
        subject_data = SubjectData(
            gender=Gender.MALE,  # all slpdb subjects were male
            age=None if age == 'x' else int(age),
            weight=None if weight == 'x' else int(weight),
        )

        yield SleepRecord(
            sleep_stages=sleep_stages,
            sleep_stage_duration=30,
            id=record_id,
            recording_start_time=start_time,
            heartbeat_times=heartbeat_times,
            subject_data=subject_data,
        )


def read_shhs(
    records_pattern: str = '*',
    use_cached_heartbeats: bool = True,
    offline: bool = False,
    keep_edfs: bool = False,
    data_dir: Optional[Union[str, Path]] = None,
) -> Iterator[SleepRecord]:
    """
    Lazily read records from SHHS (https://sleepdata.org/datasets/shhs).

    Each SHHS record consists of an `.edf` file containing raw
    polysomnography data and an `.xml` file containing annotated events.
    Since the entire SHHS dataset requires about 356 GB of disk space,
    `.edf` files can be deleted after heartbeat times have been extracted.
    Heartbeat times are cached in an `.npy` file in
    `<data_dir>/shhs/preprocessed/heartbeats`.

    Parameters
    ----------
    records_pattern : str, optional
         Glob-like pattern to select record IDs, by default `'*'`.
    use_cached_heartbeats : bool, optional
        If `True`, get heartbeat times directly from the cached `.npy`
        file, so `.edf` files are only downloaded for records which have
        not yet been preprocessed. By default `True`.
    offline : bool, optional
        If `True`, search for local files only instead of using the NSRR
        API, by default `False`.
    keep_edfs : bool, optional
        If `False`, remove `.edf` after heartbeat detection, by default
        `False`.
    data_dir : str | pathlib.Path, optional
        Directory where all datasets are stored. If `None` (default), the
        value will be taken from the configuration.

    Yields
    ------
    SleepRecord
        Each element in the generator is a :class:`SleepRecord`.
    """
    from mne.io import read_raw_edf

    DB_SLUG = 'shhs'
    ANNOTATION_DIRNAME = 'polysomnography/annotations-events-nsrr'
    EDF_DIRNAME = 'polysomnography/edfs'
    HEARTBEATS_DIRNAME = 'preprocessed/heartbeats'

    if data_dir is None:
        data_dir = get_config('data_dir')

    if not offline:
        download_url = _get_nsrr_url(DB_SLUG)

    db_dir = Path(data_dir).expanduser() / DB_SLUG
    annotations_dir = db_dir / ANNOTATION_DIRNAME
    edf_dir = db_dir / EDF_DIRNAME
    heartbeats_dir = db_dir / HEARTBEATS_DIRNAME

    for directory in (annotations_dir, edf_dir, heartbeats_dir):
        directory.mkdir(parents=True, exist_ok=True)

    if not offline:
        checksums = {}
        xml_files = _list_nsrr(
            DB_SLUG,
            ANNOTATION_DIRNAME,
            f'{records_pattern}-nsrr.xml',
            shallow=False,
        )
        checksums.update(xml_files)
        requested_records = [file[-27:-9] for file, _ in xml_files]

        edf_files = _list_nsrr(
            DB_SLUG,
            EDF_DIRNAME,
            f'{records_pattern}.edf',
            shallow=False,
        )
        checksums.update(edf_files)
    else:
        xml_files = sorted(annotations_dir.rglob(f'{records_pattern}-nsrr.xml'))
        requested_records = [str(file)[-27:-9] for file in xml_files]

    for record_id in requested_records:
        heartbeats_file = heartbeats_dir / f'{record_id}.npy'
        if use_cached_heartbeats and heartbeats_file.is_file():
            heartbeat_times = np.load(heartbeats_file)
        else:
            edf_filename = EDF_DIRNAME + f'/{record_id}.edf'
            edf_filepath = db_dir / edf_filename
            edf_was_available = edf_filepath.is_file()
            if not offline:
                _download_nsrr_file(
                    download_url + edf_filename,
                    edf_filepath,
                    checksums[edf_filename],
                )

            rec = read_raw_edf(edf_filepath, verbose=False)
            ecg = rec.get_data('ECG').ravel()
            fs = rec.info['sfreq']
            heartbeat_indices = detect_heartbeats(ecg, fs)
            heartbeat_times = heartbeat_indices / fs

            heartbeats_file.parent.mkdir(parents=True, exist_ok=True)
            np.save(heartbeats_file, heartbeat_times)

            if not edf_was_available and not keep_edfs:
                edf_filepath.unlink()

        xml_filename = ANNOTATION_DIRNAME + f'/{record_id}-nsrr.xml'
        xml_filepath = db_dir / xml_filename
        if not offline:
            _download_nsrr_file(
                download_url + xml_filename,
                xml_filepath,
                checksums[xml_filename],
            )

        parsed_xml = _parse_nsrr_xml(xml_filepath)

        yield SleepRecord(
            sleep_stages=parsed_xml.sleep_stages,
            sleep_stage_duration=parsed_xml.sleep_stage_duration,
            id=record_id[6:],  # remove subdirectory
            recording_start_time=parsed_xml.recording_start_time,
            heartbeat_times=heartbeat_times,
        )



# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""Simple interface for downloading PhysioNet data."""

import fnmatch
from pathlib import Path
from typing import Dict, Iterable, List, Optional

from tqdm import tqdm

from .utils import _download_file

_PHYSIONET_FILES_URL = 'https://physionet.org/files/'
_CHECKSUM_FILENAME = 'SHA256SUMS.txt'
_RECORDS_FILENAME = 'RECORDS'
_CHECKSUM_TYPE = 'sha256'


def _list_physionet(
    data_dir: Path,
    db_slug: str,
    db_version: Optional[str] = '1.0.0',
    pattern='*',
) -> List[str]:
    """
    List record IDs for a PhysioNet database.

    IDs can be filtered using a glob-like `pattern`.

    Parameters
    ----------
    data_dir : pathlib.Path
        Directory where all datasets are stored. Required to download the
        RECORDS-file.
    db_slug : str
        Short identifier of a database, e.g. `'mitdb'`.
    db_version : str, optional
        Version of the database, by default `'1.0.0'`.
    pattern : str, optional
        Glob-like pattern to select record IDs, by default `'*'`.

    Returns
    -------
    list[str]
        List containing record IDs as strings.
    """
    data_dir = Path(data_dir)

    records_filepath = data_dir / db_slug / _RECORDS_FILENAME
    records_url = f'{_PHYSIONET_FILES_URL}/{db_slug}/{db_version}/{_RECORDS_FILENAME}'
    checksum = _get_physionet_checksums(data_dir, db_slug, db_version)[_RECORDS_FILENAME]

    if not records_filepath.is_file():
        _download_file(records_url, records_filepath, checksum, _CHECKSUM_TYPE)

    all_records = records_filepath.read_text().splitlines()
    return fnmatch.filter(all_records, pattern)


def download_physionet(
    data_dir: Path,
    db_slug: str,
    requested_records: List[str],
    extensions: Iterable[str],
    db_version: Optional[str] = '1.0.0',
) -> None:
    """
    Download requested files from PhysioNet.

    All files with `extensions` for record IDs in `requested_records` are
    downloaded from the PhysioNet database `db_slug`.

    Parameters
    ----------
    data_dir : pathlib.Path
        Directory where all datasets are stored.
    db_slug : str
        Short identifier of a database, e.g. `'mitdb'`.
    requested_records : list[str]
        Records with those IDs are downloaded.
    extensions : Iterable[str]
        Files with those extensions are downloaded.
    db_version : str, optional
        Version of the database, by default `'1.0.0'`.
    """
    data_dir = Path(data_dir)
    checksums = _get_physionet_checksums(data_dir, db_slug, db_version)
    db_url = f'{_PHYSIONET_FILES_URL}/{db_slug}/{db_version}'

    for record_id in tqdm(requested_records, desc=f'Downloading {db_slug}'):
        for extension in extensions:
            if not extension.startswith('.'):
                extension = '.' + extension
            filepath = (data_dir / db_slug / record_id).with_suffix(extension)
            _download_file(
                f'{db_url}/{filepath.name}',
                filepath,
                checksums[filepath.name],
                checksum_type=_CHECKSUM_TYPE,
            )


def _get_physionet_checksums(
    data_dir: Path,
    db_slug: str,
    db_version: Optional[str] = '1.0.0',
) -> Dict[str, str]:
    """
    Parse PhysioNet checksums into a dictionary.

    Reads a PhysioNet checksum file and parses it into a dictionary mapping
    filenames to checksums. Tries to download the checksum file if it's not
    available on disk.

    Parameters
    ----------
    data_dir : pathlib.Path
        Directory where all datasets are stored.
    db_slug : str
        Short identifier of a database, e.g. `'mitdb'`.
    db_version : str, optional
        Version of the database, by default `'1.0.0'`.

    Returns
    -------
    dict[str, str]
        Mapping of filenames to checksums.
    """
    checksum_url = f'{_PHYSIONET_FILES_URL}/{db_slug}/{db_version}/{_CHECKSUM_FILENAME}'
    checksum_filepath = data_dir / db_slug / _CHECKSUM_FILENAME

    if not checksum_filepath.is_file():
        _download_file(checksum_url, checksum_filepath)

    checksums = {}
    for line in checksum_filepath.read_text().splitlines():
        checksum, filename = line.split()
        checksums[filename] = checksum
    return checksums



"""Functions for downloading and reading datasets."""

from .ecg_readers import ECGRecord, read_gudb, read_ltdb, read_mitdb
from .nsrr import download_nsrr, set_nsrr_token
from .physionet import download_physionet
from .sleep_readers import SleepRecord, SleepStage, read_mesa, read_shhs, read_slpdb



# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""Functions for getting and setting configuration values."""

from pathlib import Path
from typing import Any, Dict, Optional

import yaml

_DEFAULT_CONFIG_PATH = Path(__file__).parent / 'config.yml'
_USER_CONFIG_PATH = Path('~/.sleepecg/config.yml').expanduser()


def _read_yaml(path: Path) -> Dict[str, Any]:
    try:
        with open(path) as file:
            cfg = yaml.safe_load(file)
    except FileNotFoundError:
        return {}

    # empty .yml-files are loaded as `None`
    if cfg is None:
        return {}
    if not isinstance(cfg, dict):
        raise ValueError(f'Invalid YAML config file at {path}')
    return cfg


def get_config(key: Optional[str] = None) -> Any:
    """
    Read SleepECG preferences from the configuration file.

    For parameters not set in the user configuration file
    (`~/.sleepecg/config.yml`), this falls back to the default values
    defined in `site-packages/sleepecg/config.yml`. See
    :ref:`configuration` for a list of possible settings.

    Parameters
    ----------
    key : str, optional
        The configuration key to look for. If `None`, all configuration
        settings are returned in a dictionary, by default `None`.

    Returns
    -------
    typing.Any
        The configuration value.
    """
    config = _read_yaml(_DEFAULT_CONFIG_PATH)
    user_config = _read_yaml(_USER_CONFIG_PATH)
    for key in user_config:
        if key not in config:
            raise ValueError(f'Invalid key found in user config at {_USER_CONFIG_PATH}: {key}')  # noqa: E501
    config.update(user_config)

    if key is None:
        return config
    if key not in config:
        options = ', '.join(config)
        raise ValueError(f'Trying to get invalid config key: {key!r}, possible options: {options}')  # noqa: E501
    return config[key]


def set_config(**kwargs):
    """
    Set SleepECG preferences and store them to the user configuration file.

    If a value is `None`, the corresponding key is deleted from the user
    configuration. See :ref:`configuration` for a list of possible
    settings.

    Parameters
    ----------
    **kwargs: dict, optional
        The configuration keys and values to set.

    Examples
    --------
    >>> set_config(data_dir='~/.sleepecg/datasets')
    """
    default_config = _read_yaml(_DEFAULT_CONFIG_PATH)
    user_config = _read_yaml(_USER_CONFIG_PATH)

    # validate all parameters before setting anything
    for key, value in kwargs.items():
        if key not in default_config:
            options = ', '.join(default_config)
            raise ValueError(f'Trying to set invalid config key: {key!r}, possible options: {options}')  # noqa: E501

    for key, value in kwargs.items():
        if value is None:
            user_config.pop(key, None)
        else:
            user_config[key] = value

    _USER_CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
    with open(_USER_CONFIG_PATH, 'w') as user_config_file:
        yaml.dump(user_config, user_config_file)
